==> coventry-0.8.1.tar.gz: coventry-0.8.1/src/messages.cpp <==
/*
 * Copyright (C) 2021 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <moderncli/keyfile.hpp>
#include <moderncli/sync.hpp>
#include <moderncli/filesystem.hpp>

#include <unordered_set>
#include <functional>
#include <utility>
#include <thread>
#include <condition_variable>
#include <fstream>
#include <iomanip>

#include "messages.hpp"
#include "logger.hpp"
#include "events.hpp"
#include "utils.hpp"

using namespace server;

namespace {
struct args_t final {
    args_t() = default;
    args_t(keyfile keys) : config(std::move(keys)) {}
    args_t(message m) : msg(std::move(m)) {}
    args_t(const logmsg& l) : log(std::make_shared<logmsg>(l)) {}

    const std::shared_ptr<const logmsg> log;
    message msg;
    keyfile config;
};

using actor_t = void (*)(const args_t&);
using msg_t = std::pair<actor_t, args_t>;

time_t max_expires = 30;
bool stopping = false;
std::ofstream logfile, cdrfile;
std::condition_variable cvar;
unique_sync<std::list<msg_t>> msglist;

void reload(const args_t& args) {
    const auto& keys = args.config;
    auto config = keys.at("messages");
    max_expires = util::get_expires(config["expires"]);

    // Allows for log rotation...
    if(logfile.is_open())
        logfile.close();
    logfile.open(SERVER_LOGGER_PATH "/coventry.log", std::ios_base::app); // FlawFinder: ignore

    if(cdrfile.is_open())
        cdrfile.close();
    cdrfile.open(SERVER_LOGGER_PATH "/calls.log", std::ios_base::app); // FlawFinder: ignore
}

void dispatch(const args_t& args, actor_t actor) {
    sync_ptr msgs(msglist);
    if(stopping)
        return;

    msgs->emplace_back(actor, args);
    msgs.unlock();
    cvar.notify_one();
}
} // end anon namespace

message::message(std::string_view from_name, std::string_view to_uri, std::string_view route_to, std::string_view msg_subject, std::string_view msg_text, std::string_view msg_type, eXosip_t *context) : data_(std::make_shared<data>(from_name, to_uri, route_to, msg_subject, msg_text, msg_type)), ctx_(context) {
    time(&data_->expires);
    data_->expires += max_expires;
}

message::message(int dialog, int msg_status, int msg_reason, std::string_view msg_subject, std::string_view msg_text, std::string_view msg_type, eXosip_t *context) : data_(std::make_shared<data>(msg_status, msg_reason, msg_subject, msg_text, msg_type)), ctx_(context), did_(dialog) {
    time(&data_->expires);
    data_->expires += max_expires;
}


void server::logger(const std::string& msg) {
    exec::post_log(logmsg(msg));
}

void server::exec_messages(const keyfile& config) {
    auto args = args_t(config);

#ifdef DEBUG
    fsys::remove(SERVER_LOGGER_PATH "/coventry.log");
    fsys::remove(SERVER_LOGGER_PATH "/calls.log");
#endif

    reload(args);
    log_info("starting messages thread");
    while(!stopping) {
        sync_ptr msgs(msglist);
        if(!msgs->empty()) {
            auto copy = std::move(msgs->front());
            auto [actor, parms] = copy;
            msgs->pop_front();
            msgs.unlock();
            try {
                actor(parms);
            }
            catch(const std::exception& e) {
                log_error("messages actor error: {}", e.what());
            }
            continue;
        }
        cvar.wait(msgs);
    }
    log_info("stopping messages thread");
}

void exec::reload_messages(const keyfile& config) {
    dispatch(config, &reload);
}

void exec::stop_messages() {
    dispatch(args_t(), []([[maybe_unused]]const args_t& args){
        stopping = true;
    });
}

void exec::notify_local(const message& msg) {
    dispatch(msg, [](const args_t& args){
        const auto& send = args.msg;
        auto context = send.context();
        auto did = send.did();
        auto type = send.type();
        auto body = send.body();
        auto event = send.subject();
        auto status = send.status();
        auto reason = send.reason();

        debug(4, "deliver to {}", send.route());

        std::this_thread::yield();

        osip_message_t *msg = nullptr;
        const exosip::guard_t lock(context);
        auto result = eXosip_insubscription_build_notify(context, did, status, reason, &msg);
        if(msg) {
            if(!event.empty())
                osip_message_set_header(msg, "Event", event.c_str());
            osip_message_set_body(msg, body.data(), body.size());
            osip_message_set_content_type(msg, type.c_str());
            eXosip_insubscription_send_request(context, did, msg);
        }
        else
            log_error("failed to send message to {}; error={}", send.to(), result);
    });
}

void exec::deliver_local(const message& msg) {
    dispatch(msg, [](const args_t& args){
        const auto& send = args.msg;
        auto context = send.context();
        auto to = send.to();
        auto type = send.type();
        auto body = send.body();
        auto from = send.from();
        auto subject = send.subject();

        debug(4, "deliver to {}", send.route());

        std::this_thread::yield();

        osip_message_t *msg = nullptr;
        const exosip::guard_t lock(context);
        auto result = eXosip_message_build_request(context, &msg, "MESSAGE", to.c_str(), from.c_str(), send.route());
        if(msg) {
            if(!subject.empty())
                osip_message_set_subject(msg, subject.c_str());
            osip_message_set_body(msg, body.data(), body.size());
            osip_message_set_content_type(msg, type.c_str());
            allow_local(msg);
            eXosip_message_send_request(context, msg);
        }
        else
            log_error("failed to send message to {}; error={}", send.to(), result);
    });
}

void exec::post_log(const logmsg& log) {
    dispatch(log, [](const args_t& args){
        const auto& log = *args.log;
        if(!logfile.is_open())
            return;

        const struct tm now = local_time(log.created);
        logfile << std::put_time(&now, "%Y-%m-%d %X ") << log.msg << std::endl;
    });
}

void exec::post_cdr(const logmsg& cdr) {
    dispatch(cdr, [](const args_t& args){
        const auto& cdr = *args.log;
        if(!cdrfile.is_open())
            return;

        auto duration = format("{}d{}h{}m{}s", cdr.duration / 86400, (cdr.duration / 3600) % 24, (cdr.duration / 60) % 60, cdr.duration % 60);
        std::stringstream json;
        json <<  std::put_time(gmtime(&cdr.created), "{\"m\":\"cdr\",\"v\":0,\"p\":\"sip\",\"s\":\"coventry/" PROJECT_VERSION "\",\"zt\":\"%FT%TZ\",") << cdr.msg << R"(,"td":")" << duration  << "\"}";
        cdrfile << json.str() << std::endl;
    });
}

==> coventry-0.8.6.tar.gz: coventry-0.8.6/src/messages.cpp <==
/*
 * Copyright (C) 2021 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "server.hpp"
#include "messages.hpp"
#include "events.hpp"
#include "utils.hpp"

#include <moderncli/keyfile.hpp>
#include <moderncli/filesystem.hpp>
#include <moderncli/datetime.hpp>

#include <list>

using namespace server;

namespace {
struct args_t final {
    args_t() = default;
    args_t(keyfile keys) : config(std::move(keys)) {}
    args_t(message m) : msg(std::move(m)) {}
    args_t(const logmsg& l) : log(std::make_shared<logmsg>(l)) {}

    std::shared_ptr<const logmsg> log;
    message msg;
    keyfile config;
};

using actor_t = void (*)(const args_t&);
using msg_t = std::pair<actor_t, args_t>;

time_t max_expires = 30;
bool stopping = false;
std::ofstream logfile, cdrfile;
std::condition_variable cvar;
unique_sync<std::list<msg_t>> msglist;

void reload(const args_t& args) {
    const auto& keys = args.config;
    auto config = keys.at("messages");
    max_expires = util::get_expires(config["expires"]);

    // Allows for log rotation...
    if(logfile.is_open())
        logfile.close();
    logfile.open(SERVER_LOGGER_PATH "/coventry.log", std::ios_base::app); // FlawFinder: ignore

    if(cdrfile.is_open())
        cdrfile.close();
    cdrfile.open(SERVER_LOGGER_PATH "/calls.log", std::ios_base::app); // FlawFinder: ignore
}

void dispatch(const args_t& args, actor_t actor) {
    sync_ptr msgs(msglist);
    if(stopping)
        return;

    msgs->emplace_back(actor, args);
    msgs.unlock();
    cvar.notify_one();
}
} // end anon namespace

message::message(std::string_view from_name, std::string_view to_uri, std::string_view route_to, std::string_view msg_subject, std::string_view msg_text, std::string_view msg_type, eXosip_t *context) : data_(std::make_shared<data>(from_name, to_uri, route_to, msg_subject, msg_text, msg_type)), ctx_(context) {
    time(&data_->expires);
    data_->expires += max_expires;
}

message::message(int dialog, int msg_status, int msg_reason, std::string_view msg_subject, std::string_view msg_text, std::string_view msg_type, eXosip_t *context) : data_(std::make_shared<data>(msg_status, msg_reason, msg_subject, msg_text, msg_type)), ctx_(context), did_(dialog) {
    time(&data_->expires);
    data_->expires += max_expires;
}

void server::logger_notify(const std::string& msg, const char *type) {
    if(eq(type, "debug"))
        return;
    exec::post_log(logmsg(msg));
}

void server::exec_messages(const keyfile& config) {
    auto args = args_t(config);

#ifdef DEBUG
    fsys::remove(SERVER_LOGGER_PATH "/coventry.log");
    fsys::remove(SERVER_LOGGER_PATH "/calls.log");
#endif

    reload(args);
    logger.info("starting messages thread");
    while(!stopping) {
        sync_ptr msgs(msglist);
        if(!msgs->empty()) {
            auto copy = std::move(msgs->front());
            auto [actor, parms] = copy;
            msgs->pop_front();
            msgs.unlock();
            try {
                actor(parms);
            }
            catch(const std::exception& e) {
                logger.error("messages actor error: {}", e.what());
            }
            continue;
        }
        cvar.wait(msgs);
    }
    logger.info("stopping messages thread");
}

void exec::reload_messages(const keyfile& config) {
    dispatch(config, &reload);
}

void exec::stop_messages() {
    dispatch(args_t(), []([[maybe_unused]]const args_t& args){
        stopping = true;
    });
}

void exec::notify_local(const message& msg) {
    dispatch(msg, [](const args_t& args){
        const auto& send = args.msg;
        auto context = send.context();
        auto did = send.did();
        auto type = send.type();
        auto body = send.body();
        auto event = send.subject();
        auto status = send.status();
        auto reason = send.reason();

        logger.debug(4, "deliver to {}", send.route());

        std::this_thread::yield();

        osip_message_t *msg = nullptr;
        const exosip::guard_t lock(context);
        auto result = eXosip_insubscription_build_notify(context, did, status, reason, &msg);
        if(msg) {
            if(!event.empty())
                osip_message_set_header(msg, "Event", event.c_str());
            osip_message_set_body(msg, body.data(), body.size());
            osip_message_set_content_type(msg, type.c_str());
            eXosip_insubscription_send_request(context, did, msg);
        }
        else
            logger.error("failed to send message to {}; error={}", send.to(), result);
    });
}

void exec::deliver_local(const message& msg) {
    dispatch(msg, [](const args_t& args){
        const auto& send = args.msg;
        auto context = send.context();
        auto to = send.to();
        auto type = send.type();
        auto body = send.body();
        auto from = send.from();
        auto subject = send.subject();

        logger.debug(4, "deliver to {}", send.route());

        std::this_thread::yield();

        osip_message_t *msg = nullptr;
        const exosip::guard_t lock(context);
        auto result = eXosip_message_build_request(context, &msg, "MESSAGE", to.c_str(), from.c_str(), send.route());
        if(msg) {
            if(!subject.empty())
                osip_message_set_subject(msg, subject.c_str());
            osip_message_set_body(msg, body.data(), body.size());
            osip_message_set_content_type(msg, type.c_str());
            allow_local(msg);
            eXosip_message_send_request(context, msg);
        }
        else
            logger.error("failed to send message to {}; error={}", send.to(), result);
    });
}

void exec::post_log(const logmsg& log) {
    dispatch(log, [](const args_t& args){
        const auto& log = *args.log;
        if(!logfile.is_open())
            return;

        const struct tm now = local_time(log.created);
        logfile << std::put_time(&now, ISO_DATETIME) << " " << log.msg << std::endl;
    });
}

void exec::post_cdr(const logmsg& cdr) {
    dispatch(cdr, [](const args_t& args){
        const auto& cdr = *args.log;
        if(!cdrfile.is_open())
            return;

        auto duration = format("{}d{}h{}m{}s", cdr.duration / 86400, (cdr.duration / 3600) % 24, (cdr.duration / 60) % 60, cdr.duration % 60);
        std::stringstream json;
        json <<  std::put_time(gmtime(&cdr.created), "{\"m\":\"cdr\",\"v\":0,\"p\":\"sip\",\"s\":\"coventry/" PROJECT_VERSION "\",\"zt\":\"%FT%TZ\",") << cdr.msg << R"(,"td":")" << duration  << "\"}";
        cdrfile << json.str() << std::endl;
    });
}
